/**
 * A maze solver object and executable.
 *
 * Solves and displays solved mazes. Pass a list of filenames containing
 * ascii-art mazes (# for wall, ? for entrance, ! for exit). Can solve mazes
 * with one or more exit and entrance. Will show the shortest route(s) and
 * grayed-out the beginnings of attempts which would have been longer or which
 * lead to dead-ends.
 *
 * @author Jonathan Heathcote
 */
import java.util.LinkedList;
import java.util.Scanner;
import java.io.File;
public class MazeSolver
{
  // The dimensions of the maze
  private int mazeWidth;
  private int mazeHeight;
  
  // Animate the search process
  private static boolean animateSearch = false;
  
  // A 2D array which represents the maze ([x][y] addresses)
  private Cell maze[][];
  
  /**
   * Parses a maze data string and initializes instance.
   *
   * @param mazeData A string containing an Ascii maze.
   *
   * @author Jonathan Heathcote
   */
  private void parseMaze(String mazeData)
  {
    // Split the maze into lines
    String mazeDataLines[] = mazeData.split("\n");
    
    // Count the number of lines in the mazeData. Make it 2 greater than the
    // original size so that the size accounts for a one line border on top and
    // bottom which will be added later.
    mazeHeight = mazeDataLines.length + 2;
    
    // Loop through the lines of the maze to find out which line has the largest
    // number of columns.
    int maxWidth = 0;
    for (String mazeDataLine : mazeDataLines)
      if (mazeDataLine.length() > maxWidth)
        maxWidth = mazeDataLine.length();
    
    // Like above, two columns excess are added to be used as a border on either
    // side later.
    mazeWidth = maxWidth + 2;
    
    maze = new Cell[mazeWidth][mazeHeight];
    
    // Populate the maze with cells
    for (int mazeLine = 0; mazeLine < mazeHeight; mazeLine ++)
    {
      for (int mazeColumn = 0; mazeColumn < mazeWidth; mazeColumn ++)
      {
        // If this cell is on the edge it is part of the border and should
        // simply be filled with empty cells.
        if (mazeLine == 0
            || mazeColumn == 0
            || mazeLine == mazeHeight - 1
            || mazeColumn == mazeWidth - 1)
        {
          // Default to a path cell.
          maze[mazeColumn][mazeLine] = new Cell(maze,
                                                mazeColumn,mazeLine,
                                                Cell.ASCII_PATH);
        } // if
        else
        {
          // Because borders have been added the line and row numbers are off by
          // one for refrenceing the original maze data so:
          int mazeDataColumn = mazeColumn - 1;
          int mazeDataLine = mazeLine - 1;
          
          // Because not every row of the input is the correct length missing
          // columns will be padded.
          if (mazeDataColumn < mazeDataLines[mazeDataLine].length())
          {
            maze[mazeColumn][mazeLine] = new Cell(maze, mazeColumn,mazeLine,
                        mazeDataLines[mazeDataLine].charAt(mazeDataColumn));
          } // if
          else
          {
            // This is out of bounds so pad it with path (empty)
            maze[mazeColumn][mazeLine] = new Cell(maze,
                                                  mazeColumn,mazeLine,
                                                  Cell.ASCII_PATH);
          } // else
        } // else
      } // for
    } // for
  } // parseMaze
  
  /**
   * Returns a string representation of the maze.
   * 
   * This uses unicode-art and terminal escape sequences.
   * 
   * @author Jonathan Heathcote
   */
  public String toString()
  {
    String outputBuffer = "";
    
    // Skip the outer rows/columns to eliminate the border when printing
    for (int mazeLine = 1; mazeLine < mazeHeight - 1; mazeLine ++)
    {
      for (int mazeColumn = 1; mazeColumn < mazeWidth - 1; mazeColumn ++)
        outputBuffer += maze[mazeColumn][mazeLine];
      outputBuffer += "\n";
    } // for
    
    return outputBuffer;
  } // toString
  
  /**
   * Flood fill maze paths.
   * 
   * A recursive function which floods the maze paths.
   *
   * @param mazeCell A cell to start searching from.
   */
  private void floodMazePath(Cell mazeCell)
  {
    if (mazeCell.isPath())
    {
      mazeCell.setToBorder();
      if (mazeCell.leftCell() != null)
        floodMazePath(mazeCell.leftCell());
      if (mazeCell.rightCell() != null)
        floodMazePath(mazeCell.rightCell());
      if (mazeCell.upCell() != null)
        floodMazePath(mazeCell.upCell());
      if (mazeCell.downCell() != null)
        floodMazePath(mazeCell.downCell());
    } // if
  } // floodMazePath
  
  /**
   * Checks the maze for obvious mistakes.
   * 
   * Checks for inaccessible entrances and missing paths and prints warnings
   * that occur.
   */
  private void warnUserOfErrors()
  {
    // If no paths are now found then chances are the flood fill leaked into the
    // maze through a hole in the wall (or the user was just stupid and made a
    // maze with no paths...) Display a warning if this was the case.
    // While we're looping through also check to make sure that there are no
    // entrances or exits which are surrounded entirely by borders (i.e. are
    // inaccessible) and again display warnings where appropriate.
    
    // Remember if a path is found
    boolean foundPath = false;
    // Remember if a bordered entrance/exit is found
    boolean foundBlockedEntrance = false;
    boolean foundBlockedExit = false;
    // Skip the outer areas as they're just borders
    for (int mazeLine = 1; mazeLine < mazeHeight - 1; mazeLine ++)
      for (int mazeColumn = 1; mazeColumn < mazeWidth - 1; mazeColumn ++)
      {
        Cell thisIterationsCell = maze[mazeColumn][mazeLine];
        // Remember that a path has been found
        if (thisIterationsCell.isPath())
          foundPath = true;
        
        // Remember that a blocked entrance has been found
        if (thisIterationsCell.isEntrence()
           && thisIterationsCell.leftCell().isBorder()
           && thisIterationsCell.rightCell().isBorder()
           && thisIterationsCell.upCell().isBorder()
           && thisIterationsCell.downCell().isBorder())
          foundBlockedEntrance = true;
        
        // Remember that a blocked exit has been found
        if (thisIterationsCell.isExit()
           && thisIterationsCell.leftCell().isBorder()
           && thisIterationsCell.rightCell().isBorder()
           && thisIterationsCell.upCell().isBorder()
           && thisIterationsCell.downCell().isBorder())
          foundBlockedExit = true;
      } // for
    if (!foundPath)
      System.out.println("WARNING: There might be a hole in your maze wall or\n"
                       + "         you forgot to place any paths!");
    if (foundBlockedEntrance)
      System.out.println("WARNING: One or more of your entrences are blocked\n"
                       + "         entirely by walls or you've not made one!");
    if (foundBlockedExit)
      System.out.println("WARNING: One or more of your exits are blocked\n"
                       + "         entirely by walls or you've not made one!");
  } // warnUserOfErrors
  
  /**
   * While searching the maze compare the current cell to another.
   * 
   * Checks the specified other cell to find out if it's non-border and has not
   * been scanned or if it has been marked as further away from a starting
   * point than the current distance+1. If it has it updates that cell's
   * distance and returns a true, or if otherwise, false.
   * 
   * @param currentCell The current cell which we're searching from
   * @param otherCell Another cell to check to see if is suitable as a next step
   *   in the path. (Note: position not checked!)
   * 
   * @return true if the otherCell was a suitable next step, false if not.
   */
  private boolean useCellAsNextStep(Cell currentCell, Cell otherCell)
  {
    // A cell is suitable if it is a non-border which has not been scanned yet
    // (and so has no known distance) or has a larger distance than the next
    // step would be.
    if(!otherCell.isBorder()
       && (otherCell.getDistance() == Cell.UNDEFINED_DISTANCE
           || otherCell.getDistance() > currentCell.getDistance() + 1))
    {
      otherCell.setDistance(currentCell.getDistance() + 1);
      
      // If animations are enabled a frame will be needed now
      doAnimate();
      
      return true;
    } // if
    else
      return false;
  } // useCellAsNextStep
  
  /**
   * Highlight all the cells from here to the beginning.
   * 
   * Recersively highlights all cells.
   */
  private void highlightRouteBackToEntrance(Cell currentCell)
  {
    // If this cell is on a search path highlighted and scan the surrounding
    // cells for the previous in the searhc path.
    if (currentCell.isSearchPath())
    {
      currentCell.setHighlight(true);
      
      // If animations are enabled a frame will be needed now
      doAnimate();
      
      // Scan the surrounding cells for ones who's distance is 1 less than this
      // cell's
      Cell surroundingCells[] = { currentCell.leftCell(),
                    							currentCell.rightCell(),
                    							currentCell.upCell(),
                    							currentCell.downCell() };
      
      // Loop through the surrounding cells and if one is a match stop looking
      // any further.
      for (Cell otherCell : surroundingCells)
        if (otherCell.getDistance() == currentCell.getDistance() - 1)
        {
          highlightRouteBackToEntrance(otherCell);
          break;
        } // if
    } // if
  } // highlightRouteBackToEntrance
  
  /**
   * Solves the maze.
   * 
   * Must only be run once!
   */
  private void solveMaze()
  {
    // A queue which will hold a list of Cells to scan next. A linked list is
    // used as it provides the neccesary queue functions.
    LinkedList<Cell> searchQueue = new LinkedList<Cell>();
    
    // Populate the searchQueue with all the starting points (by looping through
    // the cells to find starting points).
    // Skip the outer areas as they're just borders
    for (int mazeLine = 1; mazeLine < mazeHeight - 1; mazeLine ++)
      for (int mazeColumn = 1; mazeColumn < mazeWidth - 1; mazeColumn ++)
        if (maze[mazeColumn][mazeLine].isEntrence())
          searchQueue.add(maze[mazeColumn][mazeLine]);
    
    // If an exit is reached then there is no point continuing scanning as any
    // other routes would be longer so when this flag is tripped the queue will
    // stop getting extended.
    boolean exitFound = false;
    
    // Work through the queue processing each item and branching/moving
    // throughout the maze. If the queue is empty then there are no more places
    // to look so stop!
    while (!searchQueue.isEmpty())
    {
      // Get the current head of the queue so we can work with it
      Cell currentCell = searchQueue.poll();
      
      // Create a list of all the surrounding cells
      Cell surroundingCells[] = { currentCell.leftCell(),
                                  currentCell.rightCell(),
                                  currentCell.upCell(),
                                  currentCell.downCell() };
      
      // Loop through the surrounding cells and see if they are suitable next
      // steps for the scan, if they are, add them to the queue
      for (Cell otherCell : surroundingCells)
        if (useCellAsNextStep(currentCell, otherCell))
        {
          // The cell is suitable and has been marked as scanned by
          // useCellAsNextStep.
          
          if (otherCell.isExit())
            exitFound = true;
          
          // If an exit has already been found then there is no point searching
          // any other directions as they would only be further so don't add it
          // to the queue. The queue is allowed to empty normally incase an
          // equally short route is to be found in the remaining queue items.
          if (!exitFound)
            searchQueue.add(otherCell);
        } // if
    } // while
    
    // Find all the exits and try to highlight the shortest routes back.
    // Skip the outer areas as they're just borders
    for (int mazeLine = 1; mazeLine < mazeHeight - 1; mazeLine ++)
      for (int mazeColumn = 1; mazeColumn < mazeWidth - 1; mazeColumn ++)
        if (maze[mazeColumn][mazeLine].isExit())
          highlightRouteBackToEntrance(maze[mazeColumn][mazeLine]);
    
  } // solveMaze
  
  /**
   * Reads and solves the maze and stores the results in the object.
   *
   * @param mazeData Ascii representation of the maze to solve formatted unix
   * style with \n newlines and a trailing empty line.
   *
   * @author Jonathan Heathcote
   */
  public MazeSolver(String mazeData)
  {
    // If animation is enabled, clear the screen
    if (animateSearch)
      System.out.print("\033[2J");
    
    // Parse the mazeData and load up the instance variables and add a 1 cell
    // border around the entire maze
    parseMaze(mazeData);
    
    // Fill the border and any empty riegeons around the maze (if it is not
    // square) starting from the top-left cell of the maze (the border which is
    // filled with path).
    floodMazePath(maze[0][0]);
    
    // Check the maze and warn the user if they've made any obvious mistakes
    warnUserOfErrors();
    
    // Solve the maze (as easy as that!)
    solveMaze();
    
  } // MazeSolver
  
  // Draws a frame onto the screen showing the current state
  private void doAnimate()
  {
    if (animateSearch)
    {
      // Move the cursor to the top-left
      System.out.print("\033[H");
      
      // Print the maze in its current state
      System.out.println(this);
      
      // Wait for next frame
      try
      {
        Thread.sleep(1);
      } // try
      catch (Exception e)
      {
        // If the user kills the animation clear the screen
        System.out.print("\033[2J\033[H");
      } ;
    } // else
  } // doAnimate
  
  public static void setAnimation(boolean value)
  {
    animateSearch = value;
  } // setAnimation
  
  /**
   * Run a demo aplication of the MazeSolver class.
   * 
   * Solves the mazes stored in the files passed as arguments.
   * 
   * @author Jonathan Heathcote
   */
  public static void main(String[] args) throws Exception
  {
    for (String arg : args)
    {
      // If the specified filename is actually an option, interpret it.
      if (arg.equals("--animate"))
      {
        Cell.setDrawNonHighlightedLines(true);
        setAnimation(true);
      } // else if
      else if (arg.equals("--no-animate"))
        setAnimation(false);
      else if (arg.equals("--clean"))
        Cell.setDrawNonHighlightedLines(false);
      else if (arg.equals("--draw-all"))
        Cell.setDrawNonHighlightedLines(true);
      else
      {
        // Create a scanner for this maze file
        Scanner fileScanner = new Scanner(new File(arg));
        
        // Read the file into a string with UNIX style newlines
        String mazeData = "";
        while(fileScanner.hasNextLine())
          mazeData += fileScanner.nextLine() + "\n";
        
        // Solve and print the maze
        System.out.println(new MazeSolver(mazeData));
      } // else
    } // for
  } // main
} // class MazeSolver
